<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads
 *
 * @package   jchoptimize/core
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */
namespace JchOptimize\Core\Html;

\defined('_JCH_EXEC') or die('Restricted access');
use JchOptimize\Core\Cdn;
use JchOptimize\Core\Exception;
use JchOptimize\Core\FeatureHelpers\DynamicJs;
use JchOptimize\Core\FileUtils;
use JchOptimize\Core\Helper;
use JchOptimize\Core\Http2Preload;
use JchOptimize\Core\Output;
use JchOptimize\Core\Url;
use JchOptimize\Platform\Paths;
use JchOptimize\Platform\Profiler;
use Joomla\DI\ContainerAwareInterface;
use Joomla\DI\ContainerAwareTrait;
use Joomla\Filesystem\File;
use Joomla\Registry\Registry;
use Joomla\Uri\Uri;
use _JchOptimizeVendor\Laminas\Cache\Storage\FlushableInterface;
use _JchOptimizeVendor\Laminas\Cache\Storage\StorageInterface;
use _JchOptimizeVendor\Laminas\EventManager\EventManager;
use _JchOptimizeVendor\Laminas\EventManager\EventManagerAwareInterface;
use _JchOptimizeVendor\Laminas\EventManager\EventManagerAwareTrait;
use _JchOptimizeVendor\Laminas\EventManager\SharedEventManagerInterface;
use function ucfirst;
/**
 *
 *
 */
class LinkBuilder implements ContainerAwareInterface, EventManagerAwareInterface
{
    use ContainerAwareTrait;
    use EventManagerAwareTrait;
    /**
     * @var Processor $oProcessor
     */
    private $oProcessor;
    /**
     * @var string cache id *
     */
    private $params;
    /**
     * @var AsyncManager
     */
    private $asyncManager;
    /**
     * @var FilesManager
     */
    private $filesManager;
    /**
     * @var StorageInterface $cache
     */
    private $cache;
    /**
     * @var Cdn
     */
    private $cdn;
    /**
     * @var Http2Preload
     */
    private $http2Preload;
    /**
     * @var FileUtils
     * @since version
     */
    private $fileUtils;
    /**
     * Constructor
     *
     * @param   Registry                     $params
     * @param   Processor                    $processor
     * @param   FilesManager                 $filesManager
     * @param   Cdn                          $cdn
     * @param   Http2Preload                 $http2Preload
     * @param   StorageInterface             $cache
     * @param   FileUtils                    $fileUtils
     * @param   SharedEventManagerInterface  $sharedEventManager
     */
    public function __construct(Registry $params, \JchOptimize\Core\Html\Processor $processor, \JchOptimize\Core\Html\FilesManager $filesManager, Cdn $cdn, Http2Preload $http2Preload, StorageInterface $cache, FileUtils $fileUtils, SharedEventManagerInterface $sharedEventManager)
    {
        $this->params = $params;
        $this->oProcessor = $processor;
        $this->filesManager = $filesManager;
        $this->cdn = $cdn;
        $this->http2Preload = $http2Preload;
        $this->cache = $cache;
        $this->fileUtils = $fileUtils;
        if (JCH_PRO) {
            $this->asyncManager = new \JchOptimize\Core\Html\AsyncManager($params);
        }
        $this->setEventManager(new EventManager($sharedEventManager));
    }
    public function prependChildToHead($sChild)
    {
        $sHeadHtml = \preg_replace('#<head[^>]*+>#i', '<head>' . "\n\t" . $sChild, $this->oProcessor->getHeadHtml(), 1);
        $this->oProcessor->setHeadHtml($sHeadHtml);
    }
    public function appendChildToHead($sChild, $bCleanReplacement = \false)
    {
        if ($bCleanReplacement) {
            $sChild = Helper::cleanReplacement($sChild);
        }
        $sHeadHtml = $this->oProcessor->getHeadHtml();
        $sHeadHtml = \preg_replace('#' . \JchOptimize\Core\Html\Parser::HTML_END_HEAD_TAG() . '#i', $sChild . "\n\t" . '</head>', $sHeadHtml, 1);
        $this->oProcessor->setHeadHtml($sHeadHtml);
    }
    public function addCriticalCssToHead($sCriticalCss, $id)
    {
        $sCriticalStyle = '<style id="jch-optimize-critical-css" data-id="' . $id . '">' . "\n" . $sCriticalCss . "\n" . '</style>';
        $this->appendChildToHead($sCriticalStyle, \true);
    }
    public function addExcludedJsToSection($section)
    {
        $aExcludedJs = $this->filesManager->aExcludedJs;
        //Add excluded javascript files to the bottom of the HTML section
        $sExcludedJs = \implode("\n", $aExcludedJs['ieo']) . \implode("\n", $aExcludedJs['peo']);
        $sExcludedJs = Helper::cleanReplacement($sExcludedJs);
        if ($sExcludedJs != '') {
            $this->appendChildToHTML($sExcludedJs, $section);
        }
    }
    public function appendChildToHTML($child, $section)
    {
        $sSearchArea = \preg_replace('#' . \JchOptimize\Core\Html\Parser::{'HTML_END_' . \strtoupper($section) . '_TAG'}() . '#si', "\t" . $child . "\n" . '</' . $section . '>', $this->oProcessor->getFullHtml(), 1);
        $this->oProcessor->setFullHtml($sSearchArea);
    }
    public function addDeferredJs($section)
    {
        $defers = $this->filesManager->defers;
        //If we're loading javascript dynamically add the deferred javascript files to array of files to load dynamically instead
        if ($this->params->get('pro_reduce_unused_js_enable', '0')) {
            /** @see DynamicJs::prepareJsDynamicUrls() */
            $this->container->get(DynamicJs::class)->prepareJsDynamicUrls($defers);
        } elseif (!empty($defers[0])) {
            foreach ($defers as $deferGroup) {
                foreach ($deferGroup as $deferArray) {
                    $this->appendChildToHTML($deferArray['script'], $section);
                }
            }
        }
    }
    public function setImgAttributes($aCachedImgAttributes)
    {
        $sHtml = $this->oProcessor->getBodyHtml();
        $this->oProcessor->setBodyHtml(\str_replace($this->oProcessor->images[0], $aCachedImgAttributes, $sHtml));
    }
    /**
     * Insert url of aggregated file in html
     *
     * @param   string  $id
     * @param   string  $type
     * @param   string  $section     Whether section being processed is head|body
     * @param   int     $jsLinksKey  Index key of javascript combined file
     *
     * @throws Exception\RuntimeException
     */
    public function replaceLinks(string $id, string $type, string $section = 'head', int $jsLinksKey = 0)
    {
        JCH_DEBUG ? Profiler::start('ReplaceLinks - ' . $type) : null;
        $searchArea = $this->oProcessor->getFullHtml();
        //All js files after the last excluded js will be placed at bottom of section
        if ($type == 'js' && $jsLinksKey >= $this->filesManager->jsExcludedIndex && !empty($this->filesManager->aJs[$this->filesManager->iIndex_js])) {
            $url = $this->buildUrl($id, 'js');
            //If last combined file is being inserted at the bottom of the page then
            //add the async or defer attribute
            if ($section == 'body') {
                $defer = \false;
                $async = \false;
                if ($this->params->get('loadAsynchronous', '0')) {
                    if ($this->filesManager->bLoadJsAsync) {
                        $async = \true;
                    } else {
                        $defer = \true;
                    }
                }
                //Add async attribute to last combined js file if option is set
                $newLink = $this->getNewJsLink($url, $defer, $async);
            } else {
                $newLink = $this->getNewJsLink($url);
            }
            //Insert script tag at the appropriate section in the HTML
            $searchArea = \preg_replace('#' . \JchOptimize\Core\Html\Parser::{'HTML_END_' . \strtoupper($section) . '_TAG'}() . '#si', "\t" . $newLink . "\n" . '</' . $section . '>', $searchArea, 1);
            $deferred = $this->filesManager->isFileDeferred($newLink);
            $this->http2Preload->add($url, $type, $deferred);
        } else {
            $url = $this->buildUrl($id, $type);
            $this->http2Preload->add($url, $type);
            $newLink = $this->{'getNew' . ucfirst($type) . 'Link'}($url);
            //Replace placeholders in HTML with combined files
            $searchArea = \preg_replace('#<JCH_' . \strtoupper($type) . '([^>]++)>#', $newLink, $searchArea, 1);
        }
        $this->oProcessor->setFullHtml($searchArea);
        JCH_DEBUG ? Profiler::stop('ReplaceLinks - ' . $type, \true) : null;
    }
    /**
     * Returns url of aggregated file
     *
     * @param   string  $id
     * @param   string  $type  css or js
     *
     * @return string  Url of aggregated file
     */
    public function buildUrl(string $id, string $type) : string
    {
        $htaccess = $this->params->get('htaccess', 2);
        switch ($htaccess) {
            case '1':
            case '3':
                $path = Paths::relAssetPath();
                $path = $htaccess == 3 ? $path . '3' : $path;
                $url = $path . Paths::rewriteBaseFolder() . ($this->isGz() ? 'gz' : 'nz') . '/' . $id . '.' . $type;
                break;
            case '0':
                $oUri = new Uri(Paths::relAssetPath());
                $oUri->setPath($oUri->getPath() . '2/jscss.php');
                $aVar = array();
                $aVar['f'] = $id;
                $aVar['type'] = $type;
                $aVar['gz'] = $this->isGZ() ? 'gz' : 'nz';
                $oUri->setQuery($aVar);
                $url = \htmlentities($oUri->toString());
                break;
            case '2':
            default:
                $path = Paths::cachePath();
                $url = $path . '/' . $type . '/' . $id . '.' . $type;
                // . ($this->isGz() ? '.gz' : '');
                $this->createStaticFiles($id, $type, $url);
                break;
        }
        if ($this->params->get('cookielessdomain_enable', '0') && !Url::isRootRelative($url)) {
            $url = Url::toRootRelative($url);
        }
        return $this->cdn->loadCdnResource($url);
    }
    /**
     * Check if gzip is set or enabled
     *
     * @return boolean   True if gzip parameter set and server is enabled
     */
    public function isGZ() : bool
    {
        return $this->params->get('gzip', 0) && \extension_loaded('zlib') && !\ini_get('zlib.output_compression') && \ini_get('output_handler') != 'ob_gzhandler';
    }
    /**
     * Create static combined file if not yet exists
     *
     *
     * @param   string  $id    Cache id of file
     * @param   string  $type  Type of file css|js
     * @param   string  $url   Url of combine file
     *
     * @return void
     */
    protected function createStaticFiles(string $id, string $type, string $url) : void
    {
        JCH_DEBUG ? Profiler::start('CreateStaticFiles - ' . $type) : null;
        //File path of combined file
        $combinedFile = $this->fileUtils->getPath($url);
        if (!\file_exists($combinedFile)) {
            $vars = ['f' => $id, 'type' => $type];
            $content = Output::getCombinedFile($vars, \false);
            if ($content === \false) {
                throw new Exception\RuntimeException('Error retrieving combined contents');
            }
            //Create file and any directory
            if (!File::write($combinedFile, $content)) {
                if ($this->cache instanceof FlushableInterface) {
                    $this->cache->flush();
                }
                throw new Exception\RuntimeException('Error creating static file');
            }
        }
        JCH_DEBUG ? Profiler::stop('CreateStaticFiles - ' . $type, \true) : null;
    }
    /**
     * @param $cssUrls
     */
    public function loadCssAsync($cssUrls)
    {
        if (!$this->params->get('pro_reduce_unused_css', '0')) {
            foreach ($cssUrls as $url) {
                $this->appendChildToHead($this->getPreloadStyleSheet($url, 'all'));
            }
        } else {
            $this->asyncManager->loadCssAsync($cssUrls);
        }
    }
    public function appendAsyncScriptsToHead()
    {
        if (JCH_PRO) {
            $sScript = $this->cleanScript($this->asyncManager->printHeaderScript());
            $this->appendChildToHead($sScript);
        }
    }
    /**
     *
     * @param   string  $script
     *
     * @return string
     */
    protected function cleanScript(string $script) : string
    {
        if (!Helper::isXhtml($this->oProcessor->getHtml())) {
            $script = \str_replace(array('<script type="text/javascript"><![CDATA[', '<script><![CDATA[', ']]></script>'), array('<script type="text/javascript">', '<script>', '</script>'), $script);
        }
        return $script;
    }
    public function addJsLazyLoadAssetsToHtml($id, $section)
    {
        $url = $this->buildUrl($id, 'js');
        $script = $this->getNewJsLink($url, \false, \true);
        $this->appendChildToHTML($script, $section);
    }
    /**
     * @param   string  $url  Url of file
     *
     * @return string
     */
    public function getNewCssLink(string $url) : string
    {
        //language=HTML
        return '<link rel="stylesheet" href="' . $url . '" />';
    }
    /**
     * @param   string  $url      Url of file
     * @param   bool    $isDefer  If true the 'defer attribute will be added to the script element
     * @param   bool    $isASync  If true the 'async' attribute will be added to the script element
     *
     * @return string
     */
    public function getNewJsLink(string $url, bool $isDefer = \false, bool $isASync = \false) : string
    {
        $deferAttr = $isDefer ? $this->getFormattedHtmlAttribute('defer') : '';
        $asyncAttr = $isASync ? $this->getFormattedHtmlAttribute('async') : '';
        return '<script src="' . $url . '"' . $asyncAttr . $deferAttr . '></script>';
    }
    public function getPreloadLink(array $attr) : string
    {
        $crossorigin = !empty($attr['crossorigin']) ? ' crossorigin' : '';
        $media = !empty($attr['media']) ? ' media="' . $attr['media'] . '"' : '';
        $type = !empty($attr['type']) ? ' type="' . $attr['type'] . '"' : '';
        $onload = !empty($attr['onload']) ? ' onload="' . $attr['onload'] . '"' : '';
        return "<link rel=\"preload\" href=\"{$attr['href']}\" as=\"{$attr['as']}\"{$type}{$media}{$crossorigin}{$onload} />";
    }
    public function getPreloadStyleSheet(string $url, string $media) : string
    {
        $attr = ['as' => 'style', 'onload' => 'this.rel=\'stylesheet\'', 'href' => $url, 'media' => $media];
        return $this->getPreloadLink($attr);
    }
    public function getPreconnectLink($domain) : string
    {
        //language=HTML
        return '<link rel="preconnect" href="' . $domain . '" crossorigin />';
    }
    public function getModulePreloadLink(string $url) : string
    {
        //language=HTML
        return '<link rel="modulepreload" href="' . $url . '" />';
    }
    /**
     * Returns HTML attribute properly formatted for XHTML/XML or HTML5
     *
     * @param   string  $attr
     *
     * @return string
     */
    private function getFormattedHtmlAttribute(string $attr) : string
    {
        return Helper::isXhtml($this->oProcessor->getHtml()) ? ' ' . $attr . '="' . $attr . '"' : ' ' . $attr;
    }
    public function preProcessHtml()
    {
        JCH_DEBUG ? Profiler::start('PreProcessHtml') : null;
        $this->getEventManager()->trigger(__FUNCTION__, $this);
        JCH_DEBUG ? Profiler::start('PreProcessHtml', \true) : null;
    }
    public function postProcessHtml()
    {
        JCH_DEBUG ? Profiler::start('PostProcessHtml') : null;
        $this->getEventManager()->trigger(__FUNCTION__, $this);
        JCH_DEBUG ? Profiler::stop('PostProcessHtml', \true) : null;
    }
}
